@@ -5,7 +5,7 @@
 **
 ** This file is part of the QtWaylandCompositor module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:GPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
@@ -14,24 +14,14 @@
 ** and conditions see https://www.qt.io/terms-conditions. For further
 ** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** General Public License version 3 or (at your option) any later version
+** approved by the KDE Free Qt Foundation. The licenses are as published by
+** the Free Software Foundation and appearing in the file LICENSE.GPL3
 ** included in the packaging of this file. Please review the following
 ** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -50,12 +40,16 @@
 #include <QtWaylandCompositor/QWaylandDrag>
 #endif
 #include <QtWaylandCompositor/private/qwlclientbufferintegration_p.h>
+#include <QtWaylandCompositor/private/qwaylandsurface_p.h>
+
+#if QT_CONFIG(opengl)
+#  include <QtGui/QOpenGLTexture>
+#  include <QtGui/QOpenGLFunctions>
+#endif
 
 #include <QtGui/QKeyEvent>
 #include <QtGui/QGuiApplication>
 #include <QtGui/QScreen>
-#include <QtGui/QOpenGLFunctions>
-#include <QtGui/QOpenGLTexture>
 
 #include <QtQuick/QSGSimpleTextureNode>
 #include <QtQuick/QQuickWindow>
@@ -63,7 +57,7 @@
 #include <QtCore/QMutexLocker>
 #include <QtCore/QMutex>
 
-#include <wayland-server.h>
+#include <wayland-server-core.h>
 #include <QThread>
 
 #ifndef GL_TEXTURE_EXTERNAL_OES
@@ -72,6 +66,7 @@
 
 QT_BEGIN_NAMESPACE
 
+#if QT_CONFIG(opengl)
 static const struct {
     const char * const vertexShaderSourceFile;
     const char * const fragmentShaderSourceFile;
@@ -82,7 +77,7 @@ static const struct {
     QSGMaterialType materialType;
 } bufferTypes[] = {
     // BufferFormatEgl_Null
-    { "", "", 0, 0, false, 0, {} },
+    { "", "", 0, 0, false, {}, {} },
 
     // BufferFormatEgl_RGB
     {
@@ -140,8 +135,7 @@ static const struct {
 };
 
 QWaylandBufferMaterialShader::QWaylandBufferMaterialShader(QWaylandBufferRef::BufferFormatEgl format)
-    : QSGMaterialShader()
-    , m_format(format)
+    : m_format(format)
 {
     setShaderSourceFile(QOpenGLShader::Vertex, QString::fromLatin1(bufferTypes[format].vertexShaderSourceFile));
     setShaderSourceFile(QOpenGLShader::Fragment, QString::fromLatin1(bufferTypes[format].fragmentShaderSourceFile));
@@ -163,7 +157,7 @@ void QWaylandBufferMaterialShader::updateState(const QSGMaterialShader::RenderSt
 
 const char * const *QWaylandBufferMaterialShader::attributeNames() const
 {
-    static char const *const attr[] = { "qt_VertexPosition", "qt_VertexTexCoord", 0 };
+    static char const *const attr[] = { "qt_VertexPosition", "qt_VertexTexCoord", nullptr };
     return attr;
 }
 
@@ -183,8 +177,7 @@ void QWaylandBufferMaterialShader::initialize()
 }
 
 QWaylandBufferMaterial::QWaylandBufferMaterial(QWaylandBufferRef::BufferFormatEgl format)
-    : QSGMaterial()
-    , m_format(format)
+    : m_format(format)
 {
     QOpenGLFunctions *gl = QOpenGLContext::currentContext()->functions();
 
@@ -222,9 +215,11 @@ void QWaylandBufferMaterial::bind()
     case 3:
         if (m_textures[2])
             m_textures[2]->bind(2);
+        Q_FALLTHROUGH();
     case 2:
         if (m_textures[1])
             m_textures[1]->bind(1);
+        Q_FALLTHROUGH();
     case 1:
         if (m_textures[0])
             m_textures[0]->bind(0);
@@ -258,19 +253,18 @@ void QWaylandBufferMaterial::ensureTextures(int count)
         m_textures << nullptr;
     }
 }
+#endif // QT_CONFIG(opengl)
 
-QMutex *QWaylandQuickItemPrivate::mutex = 0;
+QMutex *QWaylandQuickItemPrivate::mutex = nullptr;
 
 class QWaylandSurfaceTextureProvider : public QSGTextureProvider
 {
 public:
     QWaylandSurfaceTextureProvider()
-        : m_smooth(false)
-        , m_sgTex(0)
     {
     }
 
-    ~QWaylandSurfaceTextureProvider()
+    ~QWaylandSurfaceTextureProvider() override
     {
         if (m_sgTex)
             m_sgTex->deleteLater();
@@ -281,22 +275,29 @@ public:
         Q_ASSERT(QThread::currentThread() == thread());
         m_ref = buffer;
         delete m_sgTex;
-        m_sgTex = 0;
+        m_sgTex = nullptr;
         if (m_ref.hasBuffer()) {
             if (buffer.isSharedMemory()) {
                 m_sgTex = surfaceItem->window()->createTextureFromImage(buffer.image());
-                if (m_sgTex) {
+#if QT_CONFIG(opengl)
+                if (m_sgTex)
                     m_sgTex->bind();
-                }
+#endif
             } else {
-                QQuickWindow::CreateTextureOptions opt = QQuickWindow::TextureOwnsGLTexture;
+#if QT_CONFIG(opengl)
+                QQuickWindow::CreateTextureOptions opt;
                 QWaylandQuickSurface *surface = qobject_cast<QWaylandQuickSurface *>(surfaceItem->surface());
                 if (surface && surface->useTextureAlpha()) {
                     opt |= QQuickWindow::TextureHasAlphaChannel;
                 }
 
                 auto texture = buffer.toOpenGLTexture();
-                m_sgTex = surfaceItem->window()->createTextureFromId(texture->textureId() , QSize(surfaceItem->width(), surfaceItem->height()), opt);
+                GLuint textureId = texture->textureId();
+                auto size = surface->bufferSize();
+                m_sgTex = surfaceItem->window()->createTextureFromNativeObject(QQuickWindow::NativeObjectTexture, &textureId, 0, size, opt);
+#else
+                qCWarning(qLcWaylandCompositor) << "Without OpenGL support only shared memory textures are supported";
+#endif
             }
         }
         emit textureChanged();
@@ -311,13 +312,14 @@ public:
 
     void setSmooth(bool smooth) { m_smooth = smooth; }
 private:
-    bool m_smooth;
-    QSGTexture *m_sgTex;
+    bool m_smooth = false;
+    QSGTexture *m_sgTex = nullptr;
     QWaylandBufferRef m_ref;
 };
 
 /*!
  * \qmltype WaylandQuickItem
+ * \instantiates QWaylandQuickItem
  * \inqmlmodule QtWayland.Compositor
  * \since 5.8
  * \brief Provides a Qt Quick item that represents a WaylandView.
@@ -382,7 +384,7 @@ QWaylandQuickItem::~QWaylandQuickItem()
 QWaylandCompositor *QWaylandQuickItem::compositor() const
 {
     Q_D(const QWaylandQuickItem);
-    return d->view->surface() ? d->view->surface()->compositor() : Q_NULLPTR;
+    return d->view->surface() ? d->view->surface()->compositor() : nullptr;
 }
 
 /*!
@@ -415,7 +417,14 @@ QWaylandSurface *QWaylandQuickItem::surface() const
 void QWaylandQuickItem::setSurface(QWaylandSurface *surface)
 {
     Q_D(QWaylandQuickItem);
+    QWaylandSurface *oldSurf = d->view->surface();
+    QWaylandCompositor *oldComp = d->view->surface() ? d->view->surface()->compositor() : nullptr;
     d->view->setSurface(surface);
+    QWaylandCompositor *newComp = d->view->surface() ? d->view->surface()->compositor() : nullptr;
+    if (oldComp != newComp)
+        emit compositorChanged();
+    if (oldSurf != surface)
+        emit surfaceChanged();
     update();
 }
 
@@ -466,7 +475,7 @@ void QWaylandQuickItem::mousePressEvent(QMouseEvent *event)
         return;
     }
 
-    if (!inputRegionContains(event->pos())) {
+    if (!inputRegionContains(event->localPos())) {
         event->ignore();
         return;
     }
@@ -478,7 +487,7 @@ void QWaylandQuickItem::mousePressEvent(QMouseEvent *event)
 
     seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->localPos()), event->windowPos());
     seat->sendMousePressEvent(event->button());
-    d->hoverPos = event->pos();
+    d->hoverPos = event->localPos();
 }
 
 /*!
@@ -504,7 +513,7 @@ void QWaylandQuickItem::mouseMoveEvent(QMouseEvent *event)
 #endif // QT_CONFIG(draganddrop)
         {
             seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->localPos()), event->windowPos());
-            d->hoverPos = event->pos();
+            d->hoverPos = event->localPos();
         }
     } else {
         emit mouseMove(event->windowPos());
@@ -541,14 +550,14 @@ void QWaylandQuickItem::mouseReleaseEvent(QMouseEvent *event)
 void QWaylandQuickItem::hoverEnterEvent(QHoverEvent *event)
 {
     Q_D(QWaylandQuickItem);
-    if (!inputRegionContains(event->pos())) {
+    if (!inputRegionContains(event->posF())) {
         event->ignore();
         return;
     }
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-        seat->sendMouseMoveEvent(d->view.data(), event->pos(), mapToScene(event->pos()));
-        d->hoverPos = event->pos();
+        seat->sendMouseMoveEvent(d->view.data(), event->posF(), mapToScene(event->posF()));
+        d->hoverPos = event->posF();
     } else {
         event->ignore();
     }
@@ -561,16 +570,16 @@ void QWaylandQuickItem::hoverMoveEvent(QHoverEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (surface()) {
-        if (!inputRegionContains(event->pos())) {
+        if (!inputRegionContains(event->posF())) {
             event->ignore();
             return;
         }
     }
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-        if (event->pos() != d->hoverPos) {
-            seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->pos()), mapToScene(event->pos()));
-            d->hoverPos = event->pos();
+        if (event->posF() != d->hoverPos) {
+            seat->sendMouseMoveEvent(d->view.data(), mapToSurface(event->posF()), mapToScene(event->posF()));
+            d->hoverPos = event->posF();
         }
     } else {
         event->ignore();
@@ -585,7 +594,7 @@ void QWaylandQuickItem::hoverLeaveEvent(QHoverEvent *event)
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-        seat->setMouseFocus(Q_NULLPTR);
+        seat->setMouseFocus(nullptr);
     } else {
         event->ignore();
     }
@@ -599,13 +608,17 @@ void QWaylandQuickItem::wheelEvent(QWheelEvent *event)
 {
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
-        if (!inputRegionContains(event->pos())) {
+        if (!inputRegionContains(event->position())) {
             event->ignore();
             return;
         }
 
         QWaylandSeat *seat = compositor()->seatFor(event);
-        seat->sendMouseWheelEvent(event->orientation(), event->delta());
+        // TODO: fix this to send a single event, when diagonal scrolling is supported
+        if (event->angleDelta().x() != 0)
+            seat->sendMouseWheelEvent(Qt::Horizontal, event->angleDelta().x());
+        if (event->angleDelta().y() != 0)
+            seat->sendMouseWheelEvent(Qt::Vertical, event->angleDelta().y());
     } else {
         event->ignore();
     }
@@ -620,7 +633,10 @@ void QWaylandQuickItem::keyPressEvent(QKeyEvent *event)
     Q_D(QWaylandQuickItem);
     if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
-        seat->sendFullKeyEvent(event);
+        if (seat->setKeyboardFocus(d->view->surface()))
+            seat->sendFullKeyEvent(event);
+        else
+            qWarning() << "Unable to set keyboard focus, cannot send key press event";
     } else {
         event->ignore();
     }
@@ -649,10 +665,10 @@ void QWaylandQuickItem::touchEvent(QTouchEvent *event)
     if (d->shouldSendInputEvents() && d->touchEventsEnabled) {
         QWaylandSeat *seat = compositor()->seatFor(event);
 
-        QPoint pointPos;
+        QPointF pointPos;
         const QList<QTouchEvent::TouchPoint> &points = event->touchPoints();
         if (!points.isEmpty())
-            pointPos = points.at(0).pos().toPoint();
+            pointPos = points.at(0).pos();
 
         if (event->type() == QEvent::TouchBegin && !inputRegionContains(pointPos)) {
             event->ignore();
@@ -721,7 +737,9 @@ void QWaylandQuickItem::handleSubsurfaceAdded(QWaylandSurface *childSurface)
         childItem->setSurface(childSurface);
         childItem->setVisible(true);
         childItem->setParentItem(this);
+        childItem->setParent(this);
         connect(childSurface, &QWaylandSurface::subsurfacePositionChanged, childItem, &QWaylandQuickItem::handleSubsurfacePosition);
+        connect(childSurface, &QWaylandSurface::destroyed, childItem, &QObject::deleteLater);
     } else {
         bool success = QMetaObject::invokeMethod(d->subsurfaceHandler, "handleSubsurfaceAdded", Q_ARG(QWaylandSurface *, childSurface));
         if (!success)
@@ -729,7 +747,39 @@ void QWaylandQuickItem::handleSubsurfaceAdded(QWaylandSurface *childSurface)
     }
 }
 
+void QWaylandQuickItem::handlePlaceAbove(QWaylandSurface *referenceSurface)
+{
+    Q_D(QWaylandQuickItem);
+    auto *parent = qobject_cast<QWaylandQuickItem*>(parentItem());
+    if (!parent)
+        return;
 
+    if (parent->surface() == referenceSurface) {
+        d->placeAboveParent();
+    } else if (auto *sibling = d->findSibling(referenceSurface)) {
+        d->placeAboveSibling(sibling);
+    } else {
+        qWarning() << "Couldn't find QWaylandQuickItem for surface" << referenceSurface
+                   << "when handling wl_subsurface.place_above";
+    }
+}
+
+void QWaylandQuickItem::handlePlaceBelow(QWaylandSurface *referenceSurface)
+{
+    Q_D(QWaylandQuickItem);
+    QWaylandQuickItem *parent = qobject_cast<QWaylandQuickItem*>(parentItem());
+    if (!parent)
+        return;
+
+    if (parent->surface() == referenceSurface) {
+        d->placeBelowParent();
+    } else if (auto *sibling = d->findSibling(referenceSurface)) {
+        d->placeBelowSibling(sibling);
+    } else {
+        qWarning() << "Couldn't find QWaylandQuickItem for surface" << referenceSurface
+                   << "when handling wl_subsurface.place_below";
+    }
+}
 
 /*!
   \qmlproperty object QtWaylandCompositor::WaylandQuickItem::subsurfaceHandler
@@ -786,6 +836,14 @@ void QWaylandQuickItem::setOutput(QWaylandOutput *output)
     d->view->setOutput(output);
 }
 
+/*!
+ * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::bufferLocked
+ *
+ * This property holds whether the item's buffer is currently locked. As long as
+ * the buffer is locked, it will not be released and returned to the client.
+ *
+ * The default is false.
+ */
 /*!
  * \property QWaylandQuickItem::bufferLocked
  *
@@ -850,15 +908,17 @@ void QWaylandQuickItem::handleSurfaceChanged()
 {
     Q_D(QWaylandQuickItem);
     if (d->oldSurface) {
-        disconnect(d->oldSurface, &QWaylandSurface::hasContentChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
-        disconnect(d->oldSurface, &QWaylandSurface::parentChanged, this, &QWaylandQuickItem::parentChanged);
-        disconnect(d->oldSurface, &QWaylandSurface::sizeChanged, this, &QWaylandQuickItem::updateSize);
-        disconnect(d->oldSurface, &QWaylandSurface::bufferScaleChanged, this, &QWaylandQuickItem::updateSize);
-        disconnect(d->oldSurface, &QWaylandSurface::configure, this, &QWaylandQuickItem::updateBuffer);
-        disconnect(d->oldSurface, &QWaylandSurface::redraw, this, &QQuickItem::update);
-        disconnect(d->oldSurface, &QWaylandSurface::childAdded, this, &QWaylandQuickItem::handleSubsurfaceAdded);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::hasContentChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::parentChanged, this, &QWaylandQuickItem::parentChanged);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::destinationSizeChanged, this, &QWaylandQuickItem::updateSize);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::bufferScaleChanged, this, &QWaylandQuickItem::updateSize);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::configure, this, &QWaylandQuickItem::updateBuffer);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::redraw, this, &QQuickItem::update);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::childAdded, this, &QWaylandQuickItem::handleSubsurfaceAdded);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::subsurfacePlaceAbove, this, &QWaylandQuickItem::handlePlaceAbove);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::subsurfacePlaceBelow, this, &QWaylandQuickItem::handlePlaceBelow);
 #if QT_CONFIG(draganddrop)
-        disconnect(d->oldSurface, &QWaylandSurface::dragStarted, this, &QWaylandQuickItem::handleDragStarted);
+        disconnect(d->oldSurface.data(), &QWaylandSurface::dragStarted, this, &QWaylandQuickItem::handleDragStarted);
 #endif
 #if QT_CONFIG(im)
         disconnect(d->oldSurface->inputMethodControl(), &QWaylandInputMethodControl::updateInputMethod, this, &QWaylandQuickItem::updateInputMethod);
@@ -867,11 +927,13 @@ void QWaylandQuickItem::handleSurfaceChanged()
     if (QWaylandSurface *newSurface = d->view->surface()) {
         connect(newSurface, &QWaylandSurface::hasContentChanged, this, &QWaylandQuickItem::surfaceMappedChanged);
         connect(newSurface, &QWaylandSurface::parentChanged, this, &QWaylandQuickItem::parentChanged);
-        connect(newSurface, &QWaylandSurface::sizeChanged, this, &QWaylandQuickItem::updateSize);
+        connect(newSurface, &QWaylandSurface::destinationSizeChanged, this, &QWaylandQuickItem::updateSize);
         connect(newSurface, &QWaylandSurface::bufferScaleChanged, this, &QWaylandQuickItem::updateSize);
         connect(newSurface, &QWaylandSurface::configure, this, &QWaylandQuickItem::updateBuffer);
         connect(newSurface, &QWaylandSurface::redraw, this, &QQuickItem::update);
         connect(newSurface, &QWaylandSurface::childAdded, this, &QWaylandQuickItem::handleSubsurfaceAdded);
+        connect(newSurface, &QWaylandSurface::subsurfacePlaceAbove, this, &QWaylandQuickItem::handlePlaceAbove);
+        connect(newSurface, &QWaylandSurface::subsurfacePlaceBelow, this, &QWaylandQuickItem::handlePlaceBelow);
 #if QT_CONFIG(draganddrop)
         connect(newSurface, &QWaylandSurface::dragStarted, this, &QWaylandQuickItem::handleDragStarted);
 #endif
@@ -887,6 +949,10 @@ void QWaylandQuickItem::handleSurfaceChanged()
             QWaylandOutput *output = newSurface->compositor()->outputFor(window());
             d->view->setOutput(output);
         }
+        for (auto subsurface : QWaylandSurfacePrivate::get(newSurface)->subsurfaceChildren) {
+            if (!subsurface.isNull())
+                handleSubsurfaceAdded(subsurface.data());
+        }
 
         updateSize();
     }
@@ -947,9 +1013,14 @@ void QWaylandQuickItem::parentChanged(QWaylandSurface *newParent, QWaylandSurfac
 void QWaylandQuickItem::updateSize()
 {
     Q_D(QWaylandQuickItem);
-    if (d->sizeFollowsSurface && surface()) {
-        setSize(surface()->size() * (d->scaleFactor() / surface()->bufferScale()));
-    }
+
+    QSize size(0, 0);
+    if (surface())
+        size = surface()->destinationSize() * d->scaleFactor();
+
+    setImplicitSize(size.width(), size.height());
+    if (d->sizeFollowsSurface)
+        setSize(size);
 }
 
 /*!
@@ -989,13 +1060,31 @@ void QWaylandQuickItem::setFocusOnClick(bool focus)
  * Returns \c true if the input region of this item's surface contains the
  * position given by \a localPosition.
  */
-bool QWaylandQuickItem::inputRegionContains(const QPointF &localPosition)
+bool QWaylandQuickItem::inputRegionContains(const QPointF &localPosition) const
 {
     if (QWaylandSurface *s = surface())
-        return s->inputRegionContains(mapToSurface(localPosition).toPoint());
+        return s->inputRegionContains(mapToSurface(localPosition));
     return false;
 }
 
+// Qt 6: Remove the non-const version
+/*!
+ * Returns \c true if the input region of this item's surface contains the
+ * position given by \a localPosition.
+ */
+bool QWaylandQuickItem::inputRegionContains(const QPointF &localPosition)
+{
+    return const_cast<const QWaylandQuickItem *>(this)->inputRegionContains(localPosition);
+}
+
+/*!
+ * \qmlmethod point WaylandQuickItem::mapToSurface(point point)
+ *
+ * Maps the given \a point in this item's coordinate system to the equivalent
+ * point within the Wayland surface's coordinate system, and returns the mapped
+ * coordinate.
+ */
+
 /*!
  * Maps the given \a point in this item's coordinate system to the equivalent
  * point within the Wayland surface's coordinate system, and returns the mapped
@@ -1004,7 +1093,39 @@ bool QWaylandQuickItem::inputRegionContains(const QPointF &localPosition)
 QPointF QWaylandQuickItem::mapToSurface(const QPointF &point) const
 {
     Q_D(const QWaylandQuickItem);
-    return point / d->scaleFactor();
+    if (!surface() || surface()->destinationSize().isEmpty())
+        return point / d->scaleFactor();
+
+    qreal xScale = width() / surface()->destinationSize().width();
+    qreal yScale = height() / surface()->destinationSize().height();
+
+    return QPointF(point.x() / xScale, point.y() / yScale);
+}
+
+/*!
+ * \qmlmethod point WaylandQuickItem::mapFromSurface(point point)
+ * \since 5.13
+ *
+ * Maps the given \a point in the Wayland surfaces's coordinate system to the equivalent
+ * point within this item's coordinate system, and returns the mapped coordinate.
+ */
+
+/*!
+ * Maps the given \a point in the Wayland surfaces's coordinate system to the equivalent
+ * point within this item's coordinate system, and returns the mapped coordinate.
+ *
+ * \since 5.13
+ */
+QPointF QWaylandQuickItem::mapFromSurface(const QPointF &point) const
+{
+    Q_D(const QWaylandQuickItem);
+    if (!surface() || surface()->destinationSize().isEmpty())
+        return point * d->scaleFactor();
+
+    qreal xScale = width() / surface()->destinationSize().width();
+    qreal yScale = height() / surface()->destinationSize().height();
+
+    return QPointF(point.x() * xScale, point.y() * yScale);
 }
 
 /*!
@@ -1030,6 +1151,9 @@ bool QWaylandQuickItem::sizeFollowsSurface() const
     return d->sizeFollowsSurface;
 }
 
+//TODO: sizeFollowsSurface became obsolete when we added an implementation for
+//implicit size. The property is here for compatibility reasons only and should
+//be removed or at least default to false in Qt 6.
 void QWaylandQuickItem::setSizeFollowsSurface(bool sizeFollowsSurface)
 {
     Q_D(QWaylandQuickItem);
@@ -1069,9 +1193,11 @@ QVariant QWaylandQuickItem::inputMethodQuery(Qt::InputMethodQuery query, QVarian
 */
 
 /*!
-    Returns true if the item is hidden, though the texture
+    \property QWaylandQuickItem::paintEnabled
+
+    Holds \c true if the item is hidden, though the texture
     is still updated. As opposed to hiding the item by
-    setting \l{Item::visible}{visible} to \c false, setting this property to \c false
+    setting \l{QQuickItem::}{visible} to \c false, setting this property to \c false
     will not prevent mouse or keyboard input from reaching item.
 */
 bool QWaylandQuickItem::paintEnabled() const
@@ -1083,10 +1209,28 @@ bool QWaylandQuickItem::paintEnabled() const
 void QWaylandQuickItem::setPaintEnabled(bool enabled)
 {
     Q_D(QWaylandQuickItem);
-    d->paintEnabled = enabled;
+
+    if (enabled != d->paintEnabled) {
+        d->paintEnabled = enabled;
+        emit paintEnabledChanged();
+    }
+
     update();
 }
 
+/*!
+    \qmlproperty  bool QtWaylandCompositor::WaylandQuickItem::touchEventsEnabled
+
+    This property holds \c true if touch events are forwarded to the client
+    surface, \c false otherwise.
+*/
+
+/*!
+    \property QWaylandQuickItem::touchEventsEnabled
+
+    This property holds \c true if touch events are forwarded to the client
+    surface, \c false otherwise.
+*/
 bool QWaylandQuickItem::touchEventsEnabled() const
 {
     Q_D(const QWaylandQuickItem);
@@ -1113,12 +1257,14 @@ void QWaylandQuickItem::updateWindow()
 
     if (d->connectedWindow) {
         disconnect(d->connectedWindow, &QQuickWindow::beforeSynchronizing, this, &QWaylandQuickItem::beforeSync);
+        disconnect(d->connectedWindow, &QQuickWindow::screenChanged, this, &QWaylandQuickItem::updateSize);
     }
 
     d->connectedWindow = newWindow;
 
     if (d->connectedWindow) {
         connect(d->connectedWindow, &QQuickWindow::beforeSynchronizing, this, &QWaylandQuickItem::beforeSync, Qt::DirectConnection);
+        connect(d->connectedWindow, &QQuickWindow::screenChanged, this, &QWaylandQuickItem::updateSize); // new screen may have new dpr
     }
 
     if (compositor() && d->connectedWindow) {
@@ -1126,6 +1272,25 @@ void QWaylandQuickItem::updateWindow()
         Q_ASSERT(output);
         d->view->setOutput(output);
     }
+
+    updateSize(); // because scaleFactor depends on devicePixelRatio, which may be different for the new window
+}
+
+void QWaylandQuickItem::updateOutput()
+{
+    Q_D(QWaylandQuickItem);
+    if (d->view->output() == d->connectedOutput)
+        return;
+
+    if (d->connectedOutput)
+        disconnect(d->connectedOutput, &QWaylandOutput::scaleFactorChanged, this, &QWaylandQuickItem::updateSize);
+
+    d->connectedOutput = d->view->output();
+
+    if (d->connectedOutput)
+        connect(d->connectedOutput, &QWaylandOutput::scaleFactorChanged, this, &QWaylandQuickItem::updateSize);
+
+    updateSize();
 }
 
 void QWaylandQuickItem::beforeSync()
@@ -1171,20 +1336,21 @@ void QWaylandQuickItem::updateInputMethod(Qt::InputMethodQueries queries)
  * If an animation is started, bufferLocked should be set to ensure the item keeps its content
  * until the animation finishes
  *
- * \sa QWaylandQuickkItem::bufferLocked
+ * \sa QWaylandQuickItem::bufferLocked
  */
 
-QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeData *)
+QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeData *data)
 {
     Q_D(QWaylandQuickItem);
+    d->lastMatrix = data->transformNode->combinedMatrix();
     const bool bufferHasContent = d->view->currentBuffer().hasContent();
 
     if (d->view->isBufferLocked() && !bufferHasContent && d->paintEnabled)
         return oldNode;
 
-    if (!bufferHasContent || !d->paintEnabled) {
+    if (!bufferHasContent || !d->paintEnabled || !surface()) {
         delete oldNode;
-        return 0;
+        return nullptr;
     }
 
     QWaylandBufferRef ref = d->view->currentBuffer();
@@ -1192,7 +1358,11 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
     const QRectF rect = invertY ? QRectF(0, height(), width(), -height())
                                 : QRectF(0, 0, width(), height());
 
-    if (ref.isSharedMemory() || bufferTypes[ref.bufferFormatEgl()].canProvideTexture) {
+    if (ref.isSharedMemory()
+#if QT_CONFIG(opengl)
+            || bufferTypes[ref.bufferFormatEgl()].canProvideTexture
+#endif
+    ) {
         // This case could covered by the more general path below, but this is more efficient (especially when using ShaderEffect items).
         QSGSimpleTextureNode *node = static_cast<QSGSimpleTextureNode *>(oldNode);
 
@@ -1213,46 +1383,53 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
         d->provider->setSmooth(smooth());
         node->setRect(rect);
 
-        return node;
-    } else {
-        Q_ASSERT(!d->provider);
-
-        QSGGeometryNode *node = static_cast<QSGGeometryNode *>(oldNode);
-
-        if (!node) {
-            node = new QSGGeometryNode;
-            d->newTexture = true;
-        }
-
-        QSGGeometry *geometry = node->geometry();
-        QWaylandBufferMaterial *material = static_cast<QWaylandBufferMaterial *>(node->material());
-
-        if (!geometry)
-            geometry = new QSGGeometry(QSGGeometry::defaultAttributes_TexturedPoint2D(), 4);
-
-        if (!material)
-            material = new QWaylandBufferMaterial(ref.bufferFormatEgl());
-
-        if (d->newTexture) {
-            d->newTexture = false;
-            for (int plane = 0; plane < bufferTypes[ref.bufferFormatEgl()].planeCount; plane++)
-                if (auto texture = ref.toOpenGLTexture(plane))
-                    material->setTextureForPlane(plane, texture);
-            material->bind();
-        }
-
-        QSGGeometry::updateTexturedRectGeometry(geometry, rect, QRectF(0, 0, 1, 1));
-
-        node->setGeometry(geometry);
-        node->setFlag(QSGNode::OwnsGeometry, true);
-
-        node->setMaterial(material);
-        node->setFlag(QSGNode::OwnsMaterial, true);
+        qreal scale = surface()->bufferScale();
+        QRectF source = surface()->sourceGeometry();
+        node->setSourceRect(QRectF(source.topLeft() * scale, source.size() * scale));
 
         return node;
     }
 
-    Q_UNREACHABLE();
+#if QT_CONFIG(opengl)
+    Q_ASSERT(!d->provider);
+
+    QSGGeometryNode *node = static_cast<QSGGeometryNode *>(oldNode);
+
+    if (!node) {
+        node = new QSGGeometryNode;
+        d->newTexture = true;
+    }
+
+    QSGGeometry *geometry = node->geometry();
+    QWaylandBufferMaterial *material = static_cast<QWaylandBufferMaterial *>(node->material());
+
+    if (!geometry)
+        geometry = new QSGGeometry(QSGGeometry::defaultAttributes_TexturedPoint2D(), 4);
+
+    if (!material)
+        material = new QWaylandBufferMaterial(ref.bufferFormatEgl());
+
+    if (d->newTexture) {
+        d->newTexture = false;
+        for (int plane = 0; plane < bufferTypes[ref.bufferFormatEgl()].planeCount; plane++)
+            if (auto texture = ref.toOpenGLTexture(plane))
+                material->setTextureForPlane(plane, texture);
+        material->bind();
+    }
+
+    QSGGeometry::updateTexturedRectGeometry(geometry, rect, QRectF(0, 0, 1, 1));
+
+    node->setGeometry(geometry);
+    node->setFlag(QSGNode::OwnsGeometry, true);
+
+    node->setMaterial(material);
+    node->setFlag(QSGNode::OwnsMaterial, true);
+
+    return node;
+#else
+    qCWarning(qLcWaylandCompositor) << "Without OpenGL support only shared memory textures are supported";
+    return nullptr;
+#endif // QT_CONFIG(opengl)
 }
 
 void QWaylandQuickItem::setTouchEventsEnabled(bool enabled)
@@ -1299,6 +1476,19 @@ void QWaylandQuickItem::raise()
         stackAfter(top);
 }
 
+void QWaylandQuickItem::sendMouseMoveEvent(const QPointF &position, QWaylandSeat *seat)
+{
+    if (seat == nullptr)
+        seat = compositor()->defaultSeat();
+
+    if (!seat) {
+        qWarning() << "No seat, can't send mouse event";
+        return;
+    }
+
+    seat->sendMouseMoveEvent(view(), position);
+}
+
 /*!
  * \internal
  *
@@ -1322,8 +1512,94 @@ void QWaylandQuickItem::handleDragStarted(QWaylandDrag *drag)
 
 qreal QWaylandQuickItemPrivate::scaleFactor() const
 {
-    return (view->output() ? view->output()->scaleFactor() : 1)
-            / (window ? window->devicePixelRatio() : 1);
+    qreal f = view->output() ? view->output()->scaleFactor() : 1;
+#if !defined(Q_OS_MACOS)
+    if (window)
+        f /= window->devicePixelRatio();
+#endif
+    return f;
+}
+
+QWaylandQuickItem *QWaylandQuickItemPrivate::findSibling(QWaylandSurface *surface) const
+{
+    Q_Q(const QWaylandQuickItem);
+    auto *parent = q->parentItem();
+    if (!parent)
+        return nullptr;
+
+    const auto siblings = q->parentItem()->childItems();
+    for (auto *sibling : siblings) {
+        auto *waylandItem = qobject_cast<QWaylandQuickItem *>(sibling);
+        if (waylandItem && waylandItem->surface() == surface)
+            return waylandItem;
+    }
+    return nullptr;
+}
+
+void QWaylandQuickItemPrivate::placeAboveSibling(QWaylandQuickItem *sibling)
+{
+    Q_Q(QWaylandQuickItem);
+    q->stackAfter(sibling);
+    q->setZ(sibling->z());
+    belowParent = sibling->d_func()->belowParent;
+}
+
+void QWaylandQuickItemPrivate::placeBelowSibling(QWaylandQuickItem *sibling)
+{
+    Q_Q(QWaylandQuickItem);
+    q->stackBefore(sibling);
+    q->setZ(sibling->z());
+    belowParent = sibling->d_func()->belowParent;
+}
+
+//### does not handle changes in z value if parent is a subsurface
+void QWaylandQuickItemPrivate::placeAboveParent()
+{
+    Q_Q(QWaylandQuickItem);
+    const auto siblings = q->parentItem()->childItems();
+
+    // Stack below first (bottom) sibling above parent
+    bool foundSibling = false;
+    for (auto it = siblings.cbegin(); it != siblings.cend(); ++it) {
+        QWaylandQuickItem *sibling = qobject_cast<QWaylandQuickItem*>(*it);
+        if (sibling && !sibling->d_func()->belowParent) {
+            q->stackBefore(sibling);
+            foundSibling = true;
+            break;
+        }
+    }
+
+    // No other subsurfaces above parent
+    if (!foundSibling && siblings.last() != q)
+        q->stackAfter(siblings.last());
+
+    q->setZ(q->parentItem()->z());
+    belowParent = false;
+}
+
+//### does not handle changes in z value if parent is a subsurface
+void QWaylandQuickItemPrivate::placeBelowParent()
+{
+    Q_Q(QWaylandQuickItem);
+    const auto siblings = q->parentItem()->childItems();
+
+    // Stack above last (top) sibling below parent
+    bool foundSibling = false;
+    for (auto it = siblings.crbegin(); it != siblings.crend(); ++it) {
+        QWaylandQuickItem *sibling = qobject_cast<QWaylandQuickItem*>(*it);
+        if (sibling && sibling->d_func()->belowParent) {
+            q->stackAfter(sibling);
+            foundSibling = true;
+            break;
+        }
+    }
+
+    // No other subsurfaces below parent
+    if (!foundSibling && siblings.first() != q)
+        q->stackBefore(siblings.first());
+
+    q->setZ(q->parentItem()->z() - 1.0);
+    belowParent = true;
 }
 
 QT_END_NAMESPACE
